class GitlabSyncError(ValueError):
    pass


class MovedIssueNotDefined(GitlabSyncError):
    pass


class IssueReferenceDuplicated(GitlabSyncError):
    pass


class MSProjectSyncError(ValueError):
    pass


class MSProjectValueSetError(MSProjectSyncError):
    pass


class LoadingError(MSProjectSyncError):
    pass


class ClassNotInitiated(MSProjectSyncError):
    """Tried to load a function without properly initiate the class"""



from datetime import datetime
from typing import Any
from .exceptions import MSProjectValueSetError


def raise_exception_if_not_datetime(obj) -> None:
    """
    Raise an exception if input is not of required type

    Raises: GitlabSyncError
    """
    if not isinstance(obj, datetime):
        raise MSProjectValueSetError(
            f"The object '{obj}' is of type {type(obj)} but datetime is required"
        )


def convert_to_int_or_raise_exception(value: Any) -> int:
    """
    Convert the give value to int or raise an exception of not possible

    Raises: GitlabSyncError
    """
    try:
        return int(value)
    except (ValueError, TypeError) as e:
        raise MSProjectValueSetError(
            f"Expected a value that can be converted to int, but "
            f"but value '{value}' of type {type(value)} can't be converted: {e}"
        )



# https://docs.microsoft.com/de-de/office/vba/api/project.pjsavetype
from typing import Iterable, Any, Union, Sequence

pjPromptSave = 2
pjSave = 1
pjDoNotSave = 0


class Task:
    def __getattribute__(self, item) -> Any:
        """A MS Task"""


class TasksIterator(Iterable):
    def __getitem__(self, item) -> Union["Task", Sequence["Task"]]:
        pass

    def __iter__(self):
        pass

    def Add(self) -> Task:
        return Task()


class MSProject_Project:
    """Mocking Class for
    https://docs.microsoft.com/de-de/office/vba/api/project.project
    """

    @property
    def Path(self) -> str:
        """Path for the current file (folder)"""
        return ""

    @property
    def Name(self) -> str:
        """Name of the Project"""
        return "filename.mpp"

    @property
    def Tasks(self) -> TasksIterator:
        return TasksIterator()


class COMObject_MSProject_Application:
    """Mocking Class for MS Project Application COM Object
    https://docs.microsoft.com/de-de/office/vba/api/project.application
    """

    def FileOpen(self, path: str):
        """Open a project
        https://docs.microsoft.com/en-us/previous-versions/office/developer/office-2003/aa194681(v=office.11)
        """

    @property
    def ActiveProject(self) -> MSProject_Project:
        """Get Active Project
        https://docs.microsoft.com/de-de/office/vba/api/project.application.activeproject
        """
        return MSProject_Project()

    def Quit(self, save_changes: int = pjPromptSave) -> None:
        """
        Closes the application
        https://docs.microsoft.com/de-de/office/vba/api/project.application.quit
        """

    def FileSave(self) -> bool:
        """Save active project
        https://docs.microsoft.com/de-de/office/vba/api/project.application.filesave
        """

    def FileClose(self, Save: int = pjPromptSave, NoAuto: bool = True) -> None:
        """
        Close the active project
        https://docs.microsoft.com/en-us/previous-versions/office/developer/office-2003/aa194664(v=office.11)
        """

    @property
    def Projects(self) -> Iterable[MSProject_Project]:

        return []


def Dispatch(application: str) -> COMObject_MSProject_Application:
    """Create a COM connection"""



from typing import List, Dict, Optional, overload
from logging import getLogger

import win32com.universal
from syncgitlab2msproject.custom_types import WebURL

from .exceptions import (
    MovedIssueNotDefined,
    MSProjectValueSetError,
    IssueReferenceDuplicated,
)
from .gitlab_issues import Issue
from .ms_project import Task, MSProject
from .custom_types import IssueRef

logger = getLogger(f"{__package__}.{__name__}")

GL_PREFIX = "!!DO NOT CHANGE!! Gitlab:"

DEFAULT_DURATION = 8 * 60


def get_issue_ref_id(issue: Issue) -> IssueRef:
    """
    Return the ID of an gitlab issue

    Note the
    """
    return IssueRef(issue.id)


def get_issue_web_url(issue: Issue) -> WebURL:
    """
    Get the web url from an gitlab issue
    """
    return WebURL(issue.web_url)


def set_issue_ref_to_task(task: Task, issue: Issue) -> None:
    """set reference to gitlab issues in MS Project task"""
    task.text30 = (
        f"{GL_PREFIX}{issue.id};{issue.group_id};{issue.project_id};{issue.iid}"
    )


def get_issue_ref_from_task(task: Optional[Task]) -> Optional[IssueRef]:
    """get reference to gitlab issues from MS Project task"""
    if task is not None and task.text30 and task.text30.startswith(GL_PREFIX):
        values = task.text30[len(GL_PREFIX) :].split(";")
        return IssueRef(int(values[0]))
    return None


def is_gitlab_hyperlink(url: WebURL, gitlab_url: WebURL) -> bool:
    return url.startswith(gitlab_url)


def get_weburl_from_task(task: Optional[Task], gitlab_url: WebURL) -> Optional[WebURL]:
    """
    Get the weburl from MS Project Task (is saved as hyperlink)
    """

    def check_get_url(value: Optional[str]) -> Optional[WebURL]:
        if value is not None:
            check_url = WebURL(value)
            if is_gitlab_hyperlink(check_url, gitlab_url):
                return check_url
        return None

    if task is not None:
        if (url := check_get_url(task.hyperlink_address)) is not None:
            return url
        # If not as hyperlink we also look in task.text29 field
        if (url := check_get_url(task.text29)) is not None:
            return url
    return None


def update_task_with_issue_data(
    task: Task, issue: Issue, parent_ids: Optional[List[IssueRef]] = None
) -> List[IssueRef]:
    """
    Update task with issue data

    if an issue is moved the date of the new issue is used as long it is available

    Args:
        task: The MS Project task that will be updated
        issue: the issue with the data to be considered
        parent_ids: the parent stuff

    Returns:
        list of IssueRefs that
    """
    if parent_ids is None:
        parent_ids = [get_issue_ref_id(issue)]
    else:
        parent_ids += [get_issue_ref_id(issue)]

    if (moved_ref := issue.moved_reference) is not None:
        assert moved_ref is not None
        try:
            return update_task_with_issue_data(task, moved_ref, parent_ids)
        except MovedIssueNotDefined:
            logger.warning(
                f"Issue {issue} was moved outside of context."
                f" Ignoring the issue. Please update the task {task} manually!"
            )
    else:
        set_issue_ref_to_task(task, issue)
        try:
            task.name = issue.title
            task.notes = issue.description
            if issue.due_date is not None:
                task.deadline = issue.due_date
            if issue.has_tasks or task.percent_complete == 0:
                task.percent_complete = issue.percentage_tasks_done
            task.work = int(issue.time_estimated)
            # Update duration in case it seems to be default
            if task.duration == DEFAULT_DURATION and task.estimated:
                if task.work > 0:
                    task.duration = task.work
            task.actual_work = issue.time_spent_total
            task.hyperlink_name = "Open in Gitlab"
            task.hyperlink_address = issue.web_url
            task.text29 = issue.web_url
            task.text28 = "; ".join([f'"{label}"' for label in issue.labels])
            if issue.is_closed:
                task.actual_finish = issue.closed_at
        except (MSProjectValueSetError, win32com.universal.com_error) as e:
            logger.error(
                f"FATAL: Could not sync issue {issue} to task {task}.\nError: {e}"
            )
        else:
            logger.info(f"Synced issue {issue} to task {task}")
    return parent_ids


def add_issue_as_task_to_project(tasks: MSProject, issue: Issue):
    task = tasks.add_task(issue.title)
    logger.info(f"Created {task} as it was missing for issue, now syncing it.")
    # Add a setting to allow forcing outline level on new tasks
    # task.outline_level = 1
    update_task_with_issue_data(task, issue)


class IssueFinder:
    def __init__(self, issues: List[Issue]):
        # Create Dictionary of all IDs to find moved ones and relate existing
        self.ref_id_to_issue: Dict[IssueRef, Issue] = {}
        # We also try to sync according to the weburl but only in a second step
        self.web_url_to_issue: Dict[WebURL, Issue] = {}
        for issue in issues:
            """ Set up all references to locate later on"""
            ref_id = get_issue_ref_id(issue)
            if ref_id in self.ref_id_to_issue:
                raise IssueReferenceDuplicated(
                    f"Reference ID {ref_id} was already defined! "
                    f"{self.ref_id_to_issue[ref_id]} and {issue} "
                    f"share the same Reference ID"
                )
            self.ref_id_to_issue[ref_id] = issue

            web_url = get_issue_web_url(issue)
            if web_url in self.web_url_to_issue:
                raise IssueReferenceDuplicated(
                    f"Web URL {web_url} was already defined! "
                    f"{self.web_url_to_issue[web_url]} and {issue} "
                    f"share the same Web URL"
                )
            self.web_url_to_issue[web_url] = issue

    # Overload to make mypy aware of the fact that only None is given
    # once the id is none
    @overload
    def by_ref_id(self, ref_id: IssueRef) -> Issue:
        ...

    @overload
    def by_ref_id(self, ref_id: None) -> None:
        ...

    def by_ref_id(self, ref_id: Optional[IssueRef]) -> Optional[Issue]:
        """
        Give related issue if ref_id is set and the issue is found
        If an invalid reference is given throw
        :exceptions KeyError
        """
        if ref_id is None:
            return None
        return self.ref_id_to_issue[ref_id]

    def by_web_url(self, web_url: Optional[WebURL]) -> Optional[Issue]:
        """
        Give related issue if weburl is set and the issue is found,
        If an invalid web_url is given throw
        :exceptions KeyError
        """
        if web_url is None:
            return None
        return self.web_url_to_issue[web_url]


def find_related_issue(
    task: Task, find_issue: IssueFinder, gitlab_url: WebURL
) -> Optional[Issue]:
    try:
        if (issue := find_issue.by_ref_id(get_issue_ref_from_task(task))) is not None:
            return issue
    except KeyError as key:
        logger.warning(
            f"Task {task} refers to Issue with ID {key} which was not found in ."
            f"the issues loaded from gitlab --> Ignored this reference"
        )
    try:
        if (
            issue := find_issue.by_web_url(get_weburl_from_task(task, gitlab_url))
        ) is not None:
            return issue
    except KeyError as key:
        logger.warning(
            f"Task {task} refers to Web url {key} which was not found in ."
            f"the issues loaded from gitlab --> Ignored this reference"
        )
    return None


def sync_gitlab_issues_to_ms_project(
    tasks: MSProject, issues: List[Issue], gitlab_url: WebURL
):
    """

    Args:
        tasks: MS Project Tasks that will be synchronized
        issues:  List of Gitlab Issues
        gitlab_url: the gitlab istance url to check url found in MS project against

    Returns:

    """
    ref_issue: Optional[Issue]
    # Keep track of already synced issues
    synced: List[IssueRef] = []

    # create finder
    find_issue = IssueFinder(issues)

    # Find moved issues and reference them
    non_moved: List[IssueRef] = []
    for issue in issues:
        if (ref_int_id := issue.moved_to_id) is not None:
            if (ref_issue := find_issue.by_ref_id(IssueRef(ref_int_id))) is not None:
                issue.moved_reference = ref_issue
        else:
            non_moved.append(get_issue_ref_id(issue))

    # get existing references and update them
    for task in tasks:
        if task is None:
            continue
        ref_issue = find_related_issue(task, find_issue, gitlab_url)
        if ref_issue is not None:
            logger.info(f"Syncing {ref_issue} into {task}")
            synced += update_task_with_issue_data(task, ref_issue)
        else:
            logger.info(
                f"Not Syncing {task} as a not reference to an gitlab "
                f"issue could be found"
            )
    # adding everything that was not synced and is not duplicate
    for ref_id in non_moved:
        if ref_id not in synced:
            if (ref_issue := find_issue.by_ref_id(ref_id)) is not None:
                add_issue_as_task_to_project(tasks, ref_issue)



# -*- coding: utf-8 -*-
"""
Handle the Command Line Interface
"""

import argparse
import sys
import logging
from pathlib import Path

from requests import ConnectionError
from syncgitlab2msproject import __version__, MSProject

__author__ = "Carli Freudenberg"
__copyright__ = "Carli Freudenberg"
__license__ = "mit"

from syncgitlab2msproject.custom_types import WebURL

from syncgitlab2msproject.gitlab_issues import (
    get_project_issues,
    get_gitlab_class,
    get_group_issues,
)
from syncgitlab2msproject.sync import sync_gitlab_issues_to_ms_project

_logger = logging.getLogger(f"{__package__}.{__name__}")


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(description="Sync Gitlab Issue into MS Project ")
    parser.add_argument(
        "--version",
        action="version",
        version="SyncGitlab2MSProject {ver}".format(ver=__version__),
    )

    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG,
    )

    # TODO read from ENV
    parser.add_argument(
        "--gitlab-url",
        dest="gitlab_url",
        help="URL to the gitlab instance i.e. https://gitlab.your-company.com",
        default="https://gitlab.com",
        type=str,
    )

    # TODO read from ENV
    parser.add_argument(
        "--gitlab-token",
        dest="gitlab_token",
        help="Gitlab personal access token",
        default=None,
    )

    parser.add_argument(
        "gitlab_resource_type",
        help="Gitlab resource type to sync with",
        type=str,
        choices=["project", "group"],
    )

    parser.add_argument(
        "gitlab_resource_id",
        help="Gitlab resource id to sync with",
        type=int,
    )

    parser.add_argument(
        dest="project_file",
        help="Microsoft Project File to sync with",
        type=str,
    )

    return parser.parse_args(args)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    # we are not as complicated as that we need %(name)s:
    logformat = "[%(asctime)s] %(levelname)s: %(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    args = parse_args(args)
    setup_logging(args.loglevel)
    ms_project_file = Path(args.project_file)
    if not ms_project_file.is_file():
        _logger.error(
            f"Could not open '{args.project_file}' - seems not to be a valid file."
        )
        exit(128)
    _logger.debug("Starting loading issues")

    gitlab = get_gitlab_class(args.gitlab_url, args.gitlab_token)

    if args.gitlab_resource_type == "project":
        get_issues_func = get_project_issues
    elif args.gitlab_resource_type == "group":
        get_issues_func = get_group_issues
    else:
        raise ValueError("Invalid Resource Type")

    try:
        issues = get_issues_func(gitlab, args.gitlab_resource_id)
    except ConnectionError as e:
        _logger.error(f"Error contacting gitlab instance: {e}")
        exit(64)
    else:
        with MSProject(ms_project_file.absolute()) as tasks:
            sync_gitlab_issues_to_ms_project(tasks, issues, WebURL(args.gitlab_url))
    _logger.info("Finished syncing")


def run():
    """Entry point for console_scripts"""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



from typing import Optional, List
from datetime import datetime
from logging import getLogger

import dateutil.parser
from gitlab import Gitlab

from .custom_types import GitlabIssue, GitlabUserDict
from .exceptions import MovedIssueNotDefined

logger = getLogger(f"{__package__}.{__name__}")


def get_user_identifier(user_dict: GitlabUserDict) -> str:
    """
    Return the user identifier

    keep as separate function to allow easier changes later if required
    """
    return str(user_dict["name"])


class Issue:
    """
    Wrapper class around Group/Project Issues
    """

    # The issue object itself is not dynamic only the contained obj is!
    __slots__ = ["obj", "_moved_reference"]

    def __init__(self, obj: GitlabIssue):
        self.obj: GitlabIssue = obj
        self._moved_reference: Optional[Issue] = None

    def __getattr__(self, item: str):
        """Default to get the values from the original objext"""
        return getattr(self.obj, item)

    @property
    def moved_reference(self) -> Optional["Issue"]:
        """
        get the reference to the moved issue if defined

        :exceptions MovedIssueNotDefined
        """
        if self.moved_to_id is None:
            return None
        else:
            if self._moved_reference is None:
                raise MovedIssueNotDefined(
                    "The issue is marked as moved but was not referenced "
                    "in the loaded issues, so tracking is not possible."
                )
            else:
                return self._moved_reference

    @moved_reference.setter
    def moved_reference(self, value: "Issue"):
        if not isinstance(value, Issue):
            raise ValueError("Can only set an Issue object as moved reference!")
        self._moved_reference = value

    def __str__(self):
        return f"'{self.title}' (ID: {self.id})"

    # **************************************************************
    # *** Define some default properties to allow static typing  ***
    # **************************************************************
    @property
    def id(self) -> int:
        """
        The id of an issue - it seems to be unique within an installation
        """
        return self.obj.id

    @property
    def iid(self) -> int:
        return self.obj.iid

    @property
    def project_id(self) -> int:
        return self.obj.project_id

    @property
    def group_id(self) -> int:
        return self.obj.group_id

    @property
    def has_tasks(self) -> bool:
        return self.obj.has_tasks

    @property
    def is_closed(self) -> bool:
        return str(self.obj.state).lower().strip().startswith("closed")

    @property
    def is_open(self):
        return not self.is_closed

    @property
    def percentage_tasks_done(self) -> int:
        """
        Percentage of tasks done, 0 if no tasks are defined and not closed.
        By definition always 100 if issue is closed (and not moved)

        :exceptions MovedIssueNotDefined
        """
        if self.is_closed:
            if self.moved_to_id is not None:
                # Needed for
                assert self._moved_reference is not None
                return self._moved_reference.percentage_tasks_done
            return 100
        if not self.has_tasks:
            return 0
        task = self.task_completion_status
        return round(task["completed_count"] / task["count"])

    @property
    def moved_to_id(self) -> Optional[int]:
        return self.obj.moved_to_id

    @property
    def title(self) -> str:
        return self.obj.title

    @property
    def description(self) -> str:
        return self.obj.description

    @property
    def closed_at(self) -> Optional[datetime]:
        if (val := self.obj.closed_at) is not None:
            return dateutil.parser.parse(val)
        return None

    @property
    def due_date(self) -> Optional[datetime]:
        if (val := self.obj.due_date) is not None:
            return dateutil.parser.parse(val)
        return None

    @property
    def closed_by(self) -> Optional[str]:
        if (val := self.obj.closed_by) is not None:
            return get_user_identifier(val)
        return None

    @property
    def time_estimated(self) -> float:
        """
        Time estimated in minutes
        """
        return self.obj.time_stats["time_estimate"] / 60

    @property
    def time_spent_total(self) -> float:
        """
        Total time spent in minutes
        """
        return self.obj.time_stats["total_time_spent"] / 60

    @property
    def assignees(self) -> List[str]:
        """
        list of Gitlab Assignees.

        Note in the community edition only one assignee is possible
        """
        return [get_user_identifier(user) for user in self.obj.assignees]

    @property
    def labels(self) -> List[str]:
        """
        list of labels
        """
        return self.obj.labels

    @property
    def web_url(self) -> str:
        """
        give the url from which the issue can be accessed
        """
        return self.obj.web_url


def get_gitlab_class(server: str, personal_token: Optional[str] = None) -> Gitlab:
    if personal_token is None:
        return Gitlab(server)
    else:
        return Gitlab(server, private_token=personal_token)


def get_group_issues(gitlab: Gitlab, group_id: int) -> List[Issue]:
    group = gitlab.groups.get(group_id, lazy=True)
    return [Issue(issue) for issue in group.issues.list(all=True)]


def get_project_issues(gitlab: Gitlab, project_id: int) -> List[Issue]:
    project = gitlab.projects.get(project_id, lazy=True)
    return [Issue(issue) for issue in project.issues.list(all=True)]



import functools
from typing import Callable, Optional, TypeVar

T_IN = TypeVar("T_IN")
T_OUT = TypeVar("T_OUT")

"""
Overload variants to show that the value is only None if None was given

Note: This overloading is readable but not recognized by mypy, so we drop it for
      the moment but still keep it

@overload
def make_none_safe(func: Callable[[T_IN], T_OUT]) -> Callable[[T_IN], T_OUT]: ...

@overload
def make_none_safe(func: Callable[[None], None]) -> Callable[[None], None]: ...
"""


def make_none_safe(
    func: Callable[[T_IN], T_OUT]
) -> Callable[[Optional[T_IN]], Optional[T_OUT]]:
    """
    Make sure functions will give None if the first argument is none
    """

    @functools.wraps(func)
    def none_safe(arg, *args, **kwargs):
        if arg is None:
            return None
        else:
            return func(arg, *args, **kwargs)

    return none_safe



from os import PathLike
from logging import getLogger
from typing import Sequence, Optional, Union, List, Any
from datetime import datetime
import dateutil
import pywintypes

import win32com.client

# Classes and functions to access Microsoft Project
# Inspired by https://gist.github.com/zlorb/ff122e8563793bb28f79

from win32com.universal import com_error

from .funcions import raise_exception_if_not_datetime, convert_to_int_or_raise_exception
from .exceptions import ClassNotInitiated, LoadingError, MSProjectValueSetError
from .decorators import make_none_safe
from .custom_types import ComMSProjectProject, ComMSProjectApplication

logger = getLogger(f"{__package__}.{__name__}")


@make_none_safe
def win2python_datetime(win32datetime: "pywintypes.datetime") -> datetime:
    """
    Convert MSProject time to Python time

    might be a cumbersome to convert to string first but seems the only way
    to be sure the timezone is correct
    see http://timgolden.me.uk/python/win32_how_do_i/use-a-pytime-value.html
    for an alternative.
    This solution is taken from:
    https://stackoverflow.com/questions/39028290/
    """
    return dateutil.parser.parse(str(win32datetime))


def na_win2py_datetime(win32datetime: "pywintypes.datetime") -> Optional[datetime]:
    """
    Convert also NA datetime to Python Datetype, give None if NA
    """
    if isinstance(win32datetime, str) and win32datetime.lower() == "na":
        return None
    else:
        return win2python_datetime(win32datetime)


def na_py2win_datetime(dt: Optional[datetime]) -> Union[datetime, str]:
    """
    Convert nullable datetype for usage in MS Project
    """
    if dt is None:
        return "NA"
    else:
        return dt


def get_project_path(ms_project) -> str:
    return ms_project.Path + "\\" + ms_project.Name


class MSProject(Sequence[Optional["Task"]]):
    """
    Python Wrapper around the Communication with MS Project

    Offers at task list
    """

    def __init__(self, doc_path: PathLike):
        self.project: ComMSProjectProject = None
        self._close_after: Optional[bool] = None
        self.mpp: ComMSProjectApplication = win32com.client.Dispatch(
            "MSProject.Application"
        )
        self.doc_path: PathLike = doc_path

    def __repr__(self):
        if self.project is None:
            return "<MSProject(not loaded)>"
        else:
            return f"<MSProject('{self.project.Name}')>"

    def __enter__(self):
        self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Only save on success
        if exc_type is None:
            self.save()
        if self._close_after:
            self.close()

    def _open_projects(self) -> List[str]:
        return [get_project_path(project) for project in self.mpp.Projects]

    def load(self) -> None:
        """Load a given MSProject file."""
        already_open = self._open_projects()
        try:
            self.mpp.FileOpen(str(self.doc_path))
            self.project = self.mpp.ActiveProject
            self._close_after = get_project_path(self.project) not in already_open
        except Exception as e:
            logger.exception(f"Error opening file: {self.doc_path}")
            raise LoadingError(e)

    def close(self) -> None:
        """Forces a close without saving (has to be done manually)"""
        if self.project is not None:
            try:
                self.mpp.FileClose(False)
            except com_error as e:
                logger.info(f"File close failed: {e}")
        self.mpp.Quit()
        del self.mpp

    def save(self) -> None:
        """Close an open MSProject, saving changes."""
        if self.project is not None:
            self.mpp.FileSave()

    def __len__(self) -> int:
        if self.project is None:
            raise ClassNotInitiated("Can't get length for a not loaded project")
        return self.project.Tasks.Count

    def add_task(self, name: str) -> "Task":
        ms_task = self.project.Tasks.Add(name)
        return Task(self, ms_task.ID - 1)

    def get_task(self, task_nr: int) -> Optional["Task"]:
        return self.project.Tasks(task_nr + 1)

    # TODO: Fix MYPY
    def __getitem__(self, i: int) -> Optional["Task"]:  # type: ignore
        if i >= len(self):
            raise IndexError("Out of tasks")
        if self.get_task(i) is None:
            return None
        else:
            return Task(self, i)


class Task:
    """
    Python Wrapper Class around MS Project Task API

    Find API reference here:
    https://docs.microsoft.com/office/vba/api/project.task

    this helps to use the Task object in a pythonic way, converting all values
    automatically.
    Properties naming follows PEP8 (lower case naming)
    """

    __slots__ = ("_project", "_tasknr")

    def __init__(self, project: MSProject, task_number: int):
        self._project = project
        self._tasknr = task_number

    def __repr__(self):
        return f"<Task({self._project.__repr__()}, {self._tasknr}) '{self.name}'>"

    def __str__(self):
        return f"'{self.name}' (ID: {self.id})"

    def _get_task(self):
        return self._project.get_task(self._tasknr)

    def _set_task_val(self, attribute: str, value: Any):
        """
        Set attribute to MS Project task but do not fail if set is not working
        """
        try:
            setattr(self._get_task(), attribute, value)
        except com_error as e:
            logger.error(
                f"Could not set attribute {attribute} with value '{value}' "
                f"({type(value)}) for {self}.\n Exception: '{e}'"
            )

    @property
    def id(self) -> int:
        return self._get_task().ID

    @property
    def name(self) -> str:
        return self._get_task().Name

    @name.setter
    def name(self, value: str):
        self._set_task_val("Name", value)

    @property
    def start(self) -> datetime:
        assert (val := win2python_datetime(self._get_task().Start)) is not None
        return val

    @start.setter
    def start(self, value: datetime):
        raise_exception_if_not_datetime(value)
        self._set_task_val("Start", value)

    @property
    def finish(self) -> datetime:
        assert (val := win2python_datetime(self._get_task().Finish)) is not None
        return val

    @finish.setter
    def finish(self, value: datetime):
        raise_exception_if_not_datetime(value)
        self._set_task_val("Finish", value)

    @property
    def actual_start(self) -> Optional[datetime]:
        return na_win2py_datetime(self._get_task().ActualStart)

    @actual_start.setter
    def actual_start(self, value: Optional[datetime]):
        if value is not None:
            raise_exception_if_not_datetime(value)
        self._set_task_val("ActualStart", na_py2win_datetime(value))

    @property
    def actual_finish(self) -> Optional[datetime]:
        return na_win2py_datetime(self._get_task().ActualFinish)

    @actual_finish.setter
    def actual_finish(self, value: Optional[datetime]):
        if value is not None:
            raise_exception_if_not_datetime(value)
        self._set_task_val("ActualFinish", na_py2win_datetime(value))

    @property
    def deadline(self) -> Optional[datetime]:
        return na_win2py_datetime(self._get_task().Deadline)

    @deadline.setter
    def deadline(self, value: Optional[datetime]):
        if value is not None:
            raise_exception_if_not_datetime(value)
        self._set_task_val("Deadline", na_py2win_datetime(value))

    @property
    def notes(self) -> str:
        return self._get_task().Notes

    @notes.setter
    def notes(self, value: str):
        self._set_task_val("Notes", value)

    @property
    def duration(self) -> Optional[int]:
        """Gets  the duration (in minutes) of a task.
        Read-only for summary tasks. Read/write Variant."""
        return self._get_task().duration

    @duration.setter
    def duration(self, value: int):
        self._set_task_val("duration", convert_to_int_or_raise_exception(value))

    @property
    def percent_complete(self) -> int:
        return self._get_task().PercentComplete

    @percent_complete.setter
    def percent_complete(self, value: int):
        value = convert_to_int_or_raise_exception(value)
        if isinstance(value, int) and 0 <= value <= 100:
            self._set_task_val("PercentComplete", value)
        else:
            raise MSProjectValueSetError(
                "Attribute percent_complete must be an integer between 0 and 100"
            )

    @property
    def work(self) -> int:
        """Gets or sets the work (in minutes) for the task. Read/write Variant."""
        return self._get_task().Work

    @work.setter
    def work(self, value: int):
        self._set_task_val("Work", convert_to_int_or_raise_exception(value))

    @property
    def actual_work(self):
        """
        Gets or sets the actual work (in minutes) for the task.
        Read/write Variant.
        """
        return self._get_task().ActualWork

    @actual_work.setter
    def actual_work(self, value: int):
        self._set_task_val("ActualWork", convert_to_int_or_raise_exception(value))

    @property
    def estimated(self) -> bool:
        """
        True if the task duration is an estimate.
        False if the task duration is a set value. Read/write Variant.
        """
        return self._get_task().Estimated

    @estimated.setter
    def estimated(self, value: bool):
        self._set_task_val("Estimated", value)

    @property
    def hyperlink_name(self) -> str:
        """set or get hyplerlink (name)
        see https://docs.microsoft.com/en-us/office/vba/api/project.task.hyperlink
        """
        return self._get_task().Hyperlink

    @hyperlink_name.setter
    def hyperlink_name(self, value: str):
        self._set_task_val("Hyperlink", value)

    @property
    def hyperlink_address(self) -> str:
        """set or get hyplerlink (url)
        see https://docs.microsoft.com/en-us/office/vba/api/project.task.hyperlink
        """
        return self._get_task().HyperlinkAddress

    @hyperlink_address.setter
    def hyperlink_address(self, value: str):
        self._set_task_val("HyperlinkAddress", value)

    @property
    def outline_level(self) -> int:
        return self._get_task().OutlineLevel

    @outline_level.setter
    def outline_level(self, value: int):
        value = convert_to_int_or_raise_exception(value)
        if value > 1:
            self._set_task_val("OutlineLevel", value)
        else:
            raise MSProjectValueSetError(
                "Outline Level has to be an int larger then zero."
            )

    @property
    def text1(self) -> str:
        """get or sets the Text1 Property"""
        return self._get_task().Text1

    @text1.setter
    def text1(self, value: str):
        self._set_task_val("Text1", value)

    @property
    def text2(self) -> str:
        """get or sets the Text2 Property"""
        return self._get_task().Text2

    @text2.setter
    def text2(self, value: str):
        self._set_task_val("Text2", value)

    @property
    def text3(self) -> str:
        """get or sets the Text3 Property"""
        return self._get_task().Text3

    @text3.setter
    def text3(self, value: str):
        self._set_task_val("Text3", value)

    @property
    def text4(self) -> str:
        """get or sets the Text4 Property"""
        return self._get_task().Text4

    @text4.setter
    def text4(self, value: str):
        self._set_task_val("Text4", value)

    @property
    def text5(self) -> str:
        """get or sets the Text5 Property"""
        return self._get_task().Text5

    @text5.setter
    def text5(self, value: str):
        self._set_task_val("Text5", value)

    @property
    def text6(self) -> str:
        """get or sets the Text6 Property"""
        return self._get_task().Text6

    @text6.setter
    def text6(self, value: str):
        self._set_task_val("Text6", value)

    @property
    def text7(self) -> str:
        """get or sets the Text7 Property"""
        return self._get_task().Text7

    @text7.setter
    def text7(self, value: str):
        self._set_task_val("Text7", value)

    @property
    def text8(self) -> str:
        """get or sets the Text8 Property"""
        return self._get_task().Text8

    @text8.setter
    def text8(self, value: str):
        self._set_task_val("Text8", value)

    @property
    def text9(self) -> str:
        """get or sets the Text9 Property"""
        return self._get_task().Text9

    @text9.setter
    def text9(self, value: str):
        self._set_task_val("Text9", value)

    @property
    def text10(self) -> str:
        """get or sets the Text10 Property"""
        return self._get_task().Text10

    @text10.setter
    def text10(self, value: str):
        self._set_task_val("Text10", value)

    @property
    def text11(self) -> str:
        """get or sets the Text11 Property"""
        return self._get_task().Text11

    @text11.setter
    def text11(self, value: str):
        self._set_task_val("Text11", value)

    @property
    def text12(self) -> str:
        """get or sets the Text12 Property"""
        return self._get_task().Text12

    @text12.setter
    def text12(self, value: str):
        self._set_task_val("Text12", value)

    @property
    def text13(self) -> str:
        """get or sets the Text13 Property"""
        return self._get_task().Text13

    @text13.setter
    def text13(self, value: str):
        self._set_task_val("Text13", value)

    @property
    def text14(self) -> str:
        """get or sets the Text14 Property"""
        return self._get_task().Text14

    @text14.setter
    def text14(self, value: str):
        self._set_task_val("Text14", value)

    @property
    def text15(self) -> str:
        """get or sets the Text15 Property"""
        return self._get_task().Text15

    @text15.setter
    def text15(self, value: str):
        self._set_task_val("Text15", value)

    @property
    def text16(self) -> str:
        """get or sets the Text16 Property"""
        return self._get_task().Text16

    @text16.setter
    def text16(self, value: str):
        self._set_task_val("Text16", value)

    @property
    def text17(self) -> str:
        """get or sets the Text17 Property"""
        return self._get_task().Text17

    @text17.setter
    def text17(self, value: str):
        self._set_task_val("Text17", value)

    @property
    def text18(self) -> str:
        """get or sets the Text18 Property"""
        return self._get_task().Text18

    @text18.setter
    def text18(self, value: str):
        self._set_task_val("Text18", value)

    @property
    def text19(self) -> str:
        """get or sets the Text19 Property"""
        return self._get_task().Text19

    @text19.setter
    def text19(self, value: str):
        self._set_task_val("Text19", value)

    @property
    def text20(self) -> str:
        """get or sets the Text20 Property"""
        return self._get_task().Text20

    @text20.setter
    def text20(self, value: str):
        self._set_task_val("Text20", value)

    @property
    def text21(self) -> str:
        """get or sets the Text21 Property"""
        return self._get_task().Text21

    @text21.setter
    def text21(self, value: str):
        self._set_task_val("Text21", value)

    @property
    def text22(self) -> str:
        """get or sets the Text22 Property"""
        return self._get_task().Text22

    @text22.setter
    def text22(self, value: str):
        self._set_task_val("Text22", value)

    @property
    def text23(self) -> str:
        """get or sets the Text23 Property"""
        return self._get_task().Text23

    @text23.setter
    def text23(self, value: str):
        self._set_task_val("Text23", value)

    @property
    def text24(self) -> str:
        """get or sets the Text24 Property"""
        return self._get_task().Text24

    @text24.setter
    def text24(self, value: str):
        self._set_task_val("Text24", value)

    @property
    def text25(self) -> str:
        """get or sets the Text25 Property"""
        return self._get_task().Text25

    @text25.setter
    def text25(self, value: str):
        self._set_task_val("Text25", value)

    @property
    def text26(self) -> str:
        """get or sets the Text26 Property"""
        return self._get_task().Text26

    @text26.setter
    def text26(self, value: str):
        self._set_task_val("Text26", value)

    @property
    def text27(self) -> str:
        """get or sets the Text27 Property"""
        return self._get_task().Text27

    @text27.setter
    def text27(self, value: str):
        self._set_task_val("Text27", value)

    @property
    def text28(self) -> str:
        """get or sets the Text28 Property"""
        return self._get_task().Text28

    @text28.setter
    def text28(self, value: str):
        self._set_task_val("Text28", value)

    @property
    def text29(self) -> str:
        """get or sets the Text29 Property"""
        return self._get_task().Text29

    @text29.setter
    def text29(self, value: str):
        self._set_task_val("Text29", value)

    @property
    def text30(self) -> str:
        """get or sets the Text30 Property"""
        return self._get_task().Text30

    @text30.setter
    def text30(self, value: str):
        self._set_task_val("Text30", value)



# -*- coding: utf-8 -*-
from pkg_resources import get_distribution, DistributionNotFound

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "SyncGitlab2MSProject"
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound

from .ms_project import MSProject
from .gitlab_issues import Issue
from .sync import sync_gitlab_issues_to_ms_project

__all__ = ["MSProject", "Issue", "sync_gitlab_issues_to_ms_project"]



# -*- coding: utf-8 -*-
from typing import Any, Union, Optional, Callable
from pathlib import Path
from inspect import getattr_static, signature
from datetime import datetime
import pytest

from syncgitlab2msproject.ms_project import MSProject
from syncgitlab2msproject.exceptions import (
    LoadingError,
    ClassNotInitiated,
    MSProjectSyncError,
)

__author__ = "Carli"
__copyright__ = "Carli"
__license__ = "mit"


BASE_DIR = Path(__file__).absolute().parent
TEST_FILE_NAME = "Project1.mpp"
TEST_FILE = BASE_DIR / TEST_FILE_NAME


class DoNotSave(Exception):
    pass


# TODO Set up Pytest with copying the Test file
# TODO Make sure that the copied file was not modified
#           --> https://docs.pytest.org/en/stable/fixture.html
# TODO test saving and loading


def get_prop_setter_func(obj: Any, property_name: str) -> Callable:
    return getattr_static(obj, property_name).fset


def get_prop_set_type_annotation(setter_function: Callable) -> Union[str, type]:
    """
    Get the type annotation of obj for property
    """

    # Combination of
    # https://stackoverflow.com/a/9917213/3813064 and
    # https://stackoverflow.com/questions/53949473/python-dynamically-access-type-annotation-of-a-property
    sign = signature(setter_function)
    return next(reversed(sign.parameters.values())).annotation


def test_load_manually():
    project = MSProject(TEST_FILE)
    project.load()
    print(project[0].name)
    assert TEST_FILE_NAME in repr(project[0])
    project.close()


def test_not_loaded():
    project: MSProject = MSProject(TEST_FILE)
    assert "not loaded" in repr(project)
    with pytest.raises(ClassNotInitiated):
        print(project[0].name)
    project.close()


def test_loading_wrong_file():
    with pytest.raises(LoadingError):
        with MSProject(BASE_DIR / "not_existing.mpp"):
            pass


def test_printing_all():
    assert TEST_FILE.is_file()
    print(f"Loading '{TEST_FILE}'")

    with MSProject(TEST_FILE) as tasks:
        assert TEST_FILE_NAME in repr(tasks)
        # Test defined Number of Tasks
        assert len(tasks) == 17
        # Test cycling through all tasks
        for i, task in enumerate(tasks):
            if task:
                print(f"*{i:>4}: '{task.name}'")
            else:
                print(f"#{i:>4}  >is empty<")

        task = tasks[0]

        # Test printing all
        for attribute in dir(task):
            if not attribute.startswith("_"):
                print(f"{attribute} = {getattr(task, attribute)}")


def test_setting_everything():
    def check_setting_val(obj: Any, attrib: str, prop_type: Union[type, None]):
        if prop_type == datetime:
            new_value = datetime(2020, 3, 4, 20, 15, 30).astimezone()
        elif prop_type == int:
            new_value = 1
        elif prop_type == str:
            new_value = "Test String"
        elif prop_type == bool:
            new_value = True
        elif prop_type is None:
            new_value = None
        else:
            raise ValueError(f"Invalid type {prop_type}")

        # If new val is given lets try to set it and compare it with result
        setattr(obj, attrib, new_value)
        print(f"Checking {attrib} - {prop_type} with '{new_value}'")
        if isinstance(new_value, datetime):
            # We need
            compare: datetime = getattr(obj, attrib)
            assert str(compare.astimezone()) == str(new_value)
        else:
            assert getattr(obj, attrib) == new_value

    with pytest.raises(DoNotSave):
        with MSProject(TEST_FILE) as tasks:
            # Test setting attributes based on typed annotation
            task = tasks[0]
            for attribute in dir(task):
                if not attribute.startswith("_"):
                    if setter_func := get_prop_setter_func(task, attribute):
                        prop_type = get_prop_set_type_annotation(setter_func)
                        if prop_type in (int, bool, None, datetime, str):
                            check_setting_val(task, attribute, prop_type)
                        elif prop_type == Optional[datetime]:
                            check_setting_val(task, attribute, None)
                            check_setting_val(task, attribute, datetime)
                        else:
                            print(f"{attribute} ignored {prop_type}")
            raise DoNotSave("We do not want the file to save!")


def test_percentage():
    with pytest.raises(MSProjectSyncError):
        with MSProject(TEST_FILE) as tasks:
            # Test setting attributes based on typed annotation
            task = tasks[0]
            task.percent_complete = 101



from typing import Union, Dict, NewType, Any

from gitlab.v4.objects import ProjectIssue, GroupIssue

GitlabUserDict = Dict[str, Union[str, int]]
GitlabIssue = Union[GroupIssue, ProjectIssue]
IssueRef = NewType("IssueRef", int)
# Are Dynamic Types, therefore any
ComMSProjectProject = Any
ComMSProjectApplication = Any
WebURL = NewType("WebURL", str)
